import multiprocessing
from joblib import Parallel, delayed
from .atimes_2_corr import atimes_2_corr
from .fcs2corr import Correlations
import numpy as np
from .extract_spad_photon_streams import extract_spad_photon_streams

def atimes_2_corrs_parallel(data, list_of_corr, accuracy=50, taumax="auto", perform_coarsening=True, logtau=True, split=10):
    """
    Calculate correlations between several photon streams with arrival times
    stored in macrotimes, using parallel computing to speed up the process

    Parameters
    ----------
    data : TYPE
        Object having fields det0, det1, ..., det24 which contain
        the macrotimes of the photon arrivals [in a.u.].
    list_of_corr : list
        List of correlations to calculate
        e.g. [4, 12, 'sum3', 'sum5', 'x1011'].
    accuracy : float, optional
        Accuracy with which to calculate G. The default is 50.
    taumax : float or string, optional
        Maximum tau value for which to calculate G. The default is "auto".
    perform_coarsening : Boolean, optional
        Perform coarsening. The default is True.
    logtau : Boolean, optional
        Use log spaced tau values. The default is True.
    split : float, optional
        Chunks size with which to split the data. The default is 10.

    Returns
    -------
    G : object
        object with [N x 2] matrices with tau and G values

    """
    
    if taumax == "auto":
        taumax = 1 / data.macrotime
    
    G = Correlations()
    
    Ndet = 25
    calcAv = False
    if 'av' in list_of_corr:
        # calculate the correlations of all channels and calculate average
        list_of_corr.remove('av')
        list_of_corr += list(range(Ndet))
        calcAv = True
    
    for corr in list_of_corr:
        print("Calculating correlation " + str(corr))
        
        # EXTRACT DATA
        crossCorr = False
        if type(corr) == int:
            dataExtr = getattr(data, 'det' + str(corr))
            t0 = dataExtr[:, 0]
            corrname = 'det' + str(corr)
        elif corr == "sum5" or corr == "sum3":
            print("Extracting and sorting photons")
            dataExtr = extract_spad_photon_streams(data, corr)
            t0 = dataExtr[:, 0]
            corrname = corr
        elif corr[0] == 'x':
            c0 = corr[1:3] # first channel
            c1 = corr[3:5] # second channel
            print("Extracting photons channels " + c0 + " and " + c1)
            dataExtr = getattr(data, 'det' + str(int(c0)))
            t0 = dataExtr[:, 0]
            dataExtr1 = getattr(data, 'det' + str(int(c1)))
            t1 = dataExtr1[:, 0]
            corrname = corr
            crossCorr = True
        
        # CALCULATE CORRELATIONS
        duration = t0[-1] * data.macrotime
        Nchunks = int(np.floor(duration / split))
        # go over all filters
        for j in range(np.shape(dataExtr)[1] - 1):
            print("   Filter " + str(j))
            if crossCorr == False:
                if j == 0:
                    Processed_list = Parallel(n_jobs=multiprocessing.cpu_count() - 1)(delayed(parallel_g)(t0, [1], data.macrotime, j, split, accuracy, taumax, perform_coarsening, logtau, chunk) for chunk in list(range(Nchunks)))
                else:
                    w0 = dataExtr[:, j+1]
                    Processed_list = Parallel(n_jobs=multiprocessing.cpu_count() - 1)(delayed(parallel_g)(t0, w0, data.macrotime, j, split, accuracy, taumax, perform_coarsening, logtau, chunk) for chunk in list(range(Nchunks)))
            else:
                if j == 0:
                    Processed_list = Parallel(n_jobs=multiprocessing.cpu_count() - 1)(delayed(parallel_gx)(t0, [1], t1, [1], data.macrotime, j, split, accuracy, taumax, perform_coarsening, logtau, chunk) for chunk in list(range(Nchunks)))
                else:
                    w0 = dataExtr[:, j+1]
                    w1 = dataExtr1[:, j+1]
                    Processed_list = Parallel(n_jobs=multiprocessing.cpu_count() - 1)(delayed(parallel_gx)(t0, w0, t1, w1, data.macrotime, j, split, accuracy, taumax, perform_coarsening, logtau, chunk) for chunk in list(range(Nchunks)))
            
            for chunk in range(Nchunks):
                setattr(G, corrname + "F" + str(j) + '_chunk' + str(chunk), Processed_list[chunk])
           
            # average over all chunks
            listOfFields = list(G.__dict__.keys())
            listOfFields = [i for i in listOfFields if i.startswith(corrname + "F" + str(j) + "_chunk")]
            Gav = sum(getattr(G, i) for i in listOfFields) / len(listOfFields)
            setattr(G, corrname + "F" + str(j) + '_average', Gav)
    
    if calcAv:
        # calculate average correlation of all detector elements
        for f in range(np.shape(dataExtr)[1] - 1):
            # start with correlation of detector 20 (last one)
            Gav = getattr(G, 'det' + str(Ndet-1) + 'F' + str(f) + '_average')
            # add correlations detector elements 0-19
            for det in range(Ndet - 1):
                Gav += getattr(G, 'det' + str(det) + 'F' + str(f) + '_average')
            # divide by the number of detector elements to get the average
            Gav = Gav / Ndet
            # store average in G
            setattr(G, 'F' + str(f) + '_average', Gav)
    
    return G


def parallel_g(t0, w0, macrotime, filter_number, split, accuracy, taumax, perform_coarsening, logtau, chunk):
    tstart = chunk * split / macrotime
    tstop = (chunk + 1) * split / macrotime
    tchunk = t0[(t0 >= tstart) & (t0 < tstop)]
    tchunkN = tchunk - tchunk[0]
    if filter_number == 0:
        # no filter
        Gtemp = atimes_2_corr(tchunkN, tchunkN, [1], [1], macrotime, accuracy, taumax, perform_coarsening, logtau)
    else:
        # filters
        wchunk = w0[(t0 >= tstart) & (t0 < tstop)].copy()
        Gtemp = atimes_2_corr(tchunkN, tchunkN, wchunk, wchunk, macrotime, accuracy, taumax, perform_coarsening, logtau)
    return(Gtemp)


def parallel_gx(t0, w0, t1, w1, macrotime, filter_number, split, accuracy, taumax, perform_coarsening, logtau, chunk):
    tstart = chunk * split / macrotime
    tstop = (chunk + 1) * split / macrotime
    tchunk0 = t0[(t0 >= tstart) & (t0 < tstop)]
    tchunk1 = t1[(t1 >= tstart) & (t1 < tstop)]
    # normalize time by sutracting first number
    tN = np.min([tchunk0[0], tchunk1[0]])
    tchunk0 = tchunk0 - tN
    tchunk1 = tchunk1 - tN
    if filter_number == 0:
        # no filter
        Gtemp = atimes_2_corr(tchunk0, tchunk1, [1], [1], macrotime, accuracy, taumax, perform_coarsening, logtau)
    else:
        # filters
        wchunk0 = w0[(t0 >= tstart) & (t0 < tstop)].copy()
        wchunk1 = w1[(t1 >= tstart) & (t1 < tstop)].copy()
        Gtemp = atimes_2_corr(tchunk0, tchunk1, wchunk0, wchunk1, macrotime, accuracy, taumax, perform_coarsening, logtau)
    return(Gtemp)




from matplotlib.colors import ListedColormap
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors as mcolors
from copy import deepcopy

def color_gradient(startColor, endColor, N):
    vals = np.ones((N, 4))
    vals[:, 0] = np.linspace(startColor[0], endColor[0], N)
    vals[:, 1] = np.linspace(startColor[1], endColor[1], N)
    vals[:, 2] = np.linspace(startColor[2], endColor[2], N)
    return(ListedColormap(vals))
    
    
def cmaps(color, N=256):
    switcher = {
        "blue2black": color_gradient((0, 0, 0, 1), mcolors.to_rgba('#1f77b4'), N),
        "orange2black": color_gradient((0, 0, 0, 1), mcolors.to_rgba('#ff7f0e'), N),
        "green2black": color_gradient((0, 0, 0, 1), mcolors.to_rgba('#2ca02c'), N),
        "red2black": color_gradient((0, 0, 0, 1), mcolors.to_rgba('#d62728'), N),
        "purple2black": color_gradient((0, 0, 0, 1), mcolors.to_rgba('#9467bd'), N),
    }
    # Get the function from switcher dictionary
    func = switcher.get(color, "Invalid color")
    # Execute the function
    return func


def change_color_from_map(cmap, new_color, idx=0):
    """
    Change a single color from a color map

    Parameters
    ----------
    cmap : string
        Color map (e.g. 'inferno').
    new_color : list
        list or np.array with 3 values between 0-1 (r,g,b) for the new color
    idx : int, optional
        Index of the color maps that has to be changed. The default is 0.

    Returns
    -------
    cmap : color map
        New color map with the color at idx changed to new_color.

    """
    
    if cmap == 'seaborn-colorblind':
        return cmap
    cmap = deepcopy(plt.get_cmap(cmap))
    vals = cmap.colors
    vals[idx] = new_color
    return(ListedColormap(vals))


from .restore_session import restorelib
import numpy as np


def read_ffs_file(libfile, read='active', returnObj='G'):
    """
    Read an .ffs libfile and return as an object

    Parameters
    ----------
    libfile : str
        path to the .ffs file.
    read : either 'active' or list of 4 numbers
        list indicating which image, ffs file, and correlation should be
        returned
    returnObj : str, optional
        Either 'im', 'file', 'analysis', or 'G'. The default is 'G'.
        Choose to return the image object, the ffs file object, the analysis
        object, or the correlation object, specified in 'read'

    Returns
    -------
    G
        Correlation object (by default) or image, ffs file, or analysis object.

    Example
    -------
    To return the 4rd ffs analysis from the 2nd ffs file that belongs to image 1:
        analysis = read_ffs_file(libfile, read=[0, 1, 3], returnObj='analysis')
    """
    readlist = False
    if read != 'active':
        readlist = True
    
    lib = restorelib(libfile)
    if returnObj=='ffs':
        return lib
    
    # get active image
    nIm = lib.num_images
    if readlist:
        imnr = read[0]
    else:
        imnr = lib.active_image
    if nIm < 1 or nIm <= imnr:
        return None
    im = lib.lib[imnr]
    if returnObj=='im':
        return im
    
    # get active file
    Nfiles = im.num_files
    if readlist:
        filenr = read[1]
    else:
        filenr = im.active_ffs
    if Nfiles < 1 or Nfiles <= filenr:
        return None
    file = im.ffs_list[filenr]
    if returnObj=='file':
        return file
    
    # get active analysis
    Nanal = file.num_analyses
    if readlist:
        analnr = read[2]
    else:
        analnr = file.active_analysis
    if Nanal < 1 or Nanal <= analnr:
        return None
    analysis = file.analysis_list[analnr]
    if returnObj=='analysis':
        return analysis
    
    if returnObj == 'G':
        G = analysis.corrs
        return G
    
    # get active fit
    Nfits = analysis.num_fits
    if readlist:
        fitnr = read[3]
    else:
        fitnr = analysis.active_fit
    if Nfits < 1 or Nfits <= fitnr:
        return None
    fit = analysis.fits[fitnr]

    return fit


def read_fitresults_from_ffs(libfile, read='active'):
    """
    Read fit results from a saved ffs session

    Parameters
    ----------
    libfile : path
        Path to .ffs file.
    read : str or list, optional
        Either 'active' to read the active file.
        Otherwise list with four elemetnsThe default is 'active'.

    Returns
    -------
    fitresults : TYPE
        DESCRIPTION.

    """
    fitObj = read_ffs_file(libfile, read=read, returnObj='fit')
    num_curves = len(fitObj.fit_all_curves)
    if num_curves < 1 or num_curves is None:
        return None
    fitresults = []
    for i in range(num_curves):
        is_fitted = fitObj.fit_all_curves[i].fitarray
        fitresults.append((fitObj.fit_all_curves[i].startvalues[is_fitted[0:-1]]))
    fitresults = np.asarray(fitresults)
    return fitresults


def read_g_from_ffs(libfile, read='active'):
    """
    Read an .ffs libfile and return the active correlation curves and fits

    Parameters
    ----------
    libfile : str
        path to the .ffs file.

    Returns
    -------
    G : np.array()
        2D array with for each column a correlation curve.
    tau : np.array()
        1D array with the corresponding tau values.
    Gfit : np.array()
        2D array with for each column the fitted correlation curve.
    taufit : np.array()
        1D array with the corresponding tau values.

    Note: tau and taufit may differ because the fit range is not always
    the full data range

    """
    
    analysis = read_ffs_file(libfile, read=read, returnObj='analysis')
    elements = analysis.settings.elements # central, sum3x3, sum5x5
    Ncurves = len(elements)
    
    Gsingle = analysis.get_corr(elements[0])
    N = len(Gsingle)
    
    G = np.zeros((N, Ncurves))
    Gfit = None
    tau = None
    taufit = None
    fitFound = False
    
    for i in range(Ncurves):
        Gsingle = analysis.get_corr(elements[i])
        tau = analysis.get_corr()[:,0]
        if 'crossCenterAv' in elements:
            G = Gsingle
        else:       
            G[:,i] = Gsingle[:,1]
            
        
        if len(analysis.fits) > 0 and analysis.active_fit is not None:
        
            if read == 'active':
                fitnr = analysis.active_fit
            else:
                fitnr = read[3]
            fits = analysis.fits[fitnr]
        
            fitrange = fits.fit_all_curves[0].fitrange
            start = fitrange[0]
            stop = fitrange[1]
            
            if fitFound == False:
                Gfit = np.zeros((stop-start, Ncurves))
                taufit = tau[start:stop]
                fitFound = True
        
            fitres = fits.fit_all_curves[i].fitresult
            if len(fitres) == len(Gsingle[start:stop,1]):
                Gfit[:, i] = Gsingle[start:stop,1] - fitres
        
    return G, tau, Gfit, taufit


import numpy as np
from ..tools.moving_average import moving_average
import copy


def align_hist(data, n=15):
    """
    Align lifetime histograms of different channels. Calculate moving averaged
    histograms, caculate derivative and find peak. Then shift everything
    so that the peaks are aligned.

    Parameters
    ----------
    data : object
        data object with microtime histograms.
    n : int, optional
        width of the window for the calculation of the moving average. The default is 15.

    Returns
    -------
    data : object
        same data object as input, but with histograms aligned in time.

    """
    
    histList = [i for i in list(data.__dict__.keys()) if i.startswith('hist')]
    
    hshifts = np.zeros(len(histList))
    
    i = 0
    for hist in histList:
        # get histogram
        histD = copy.deepcopy(getattr(data, hist))
        # calculate moving average
        IAv = moving_average(histD[:,1], n=n)
        # calculate derivative
        derivative = IAv[1:] - IAv[0:-1]
        # find maximum of derivative
        maxInd = np.argmax(derivative)
        # shift histogram
        histD[:,1] = np.roll(histD[:,1], -maxInd)
        # store shifted histogram in data object
        setattr(data, 'A' + hist, histD)
        # store shift value
        hshifts[i] = maxInd
        i += 1
    setattr(data, 'hshifts', hshifts)
    
    return data



import matplotlib.pyplot as plt
import numpy as np
from ..tools.bindata import bindata


def plot_intensity_traces(data, dwell_time=1, binsize=1, yscale='log', list_of_channels="all"):
    """
    Plot intensity traces from SPAD-fcs data

    Parameters
    ----------
    data : object
        Data variable, i.e. output from binfile2data.
    dwell_time : scalar, optional
        Bin time [in µs]. The default is 1.
    binsize : scalar, optional
        Number of data points in one bin. The default is 1.
    yscale : string, optional
        "linear" or "log" y scale. The default is 'log'.
    list_of_channels : list of numbers or string, optional
        List with channel numbers to be plotted, e.g. [0, 12, 24]. The default is "all".

    Returns
    -------
    h : figure
        plot with all intensity traces.

    """

    # number of channels
    Nc = np.size(data, 1)

    # channels to plot
    if list_of_channels == "all":
        list_of_channels = list(range(np.min([Nc, 25])))
    
    # bin data
    databin = bindata(data, binsize)
    # number of time points in binned data
    Nt = np.size(databin, 0)
    
    # create figure
    leg = []  # figure legend
    h = plt.figure()
    
    # bin time
    binTime = 1e-6 * dwell_time * binsize
    
    # time vector
    time = list(range(0, Nt))
    time = [i * binTime for i in time]
    
    ymax = 0
    for i in list_of_channels:
        # rescale intensity values to frequencies
        PCR = databin[:, i] / binTime
        # for lag plot in Hz, else in kHz
        scaleFactor = 1000
        if yscale == 'log':
            scaleFactor = 1
        PCRscaled = PCR / scaleFactor
        plt.plot(time, PCRscaled)
        leg.append('Pixel ' + str(i))
        ymax = np.max([ymax, np.max(PCRscaled)])
    
    plt.xlabel('Time [s]')
    plt.xlim([0, 2*time[-1] - time[-2]])
    if yscale == 'linear':
        plt.ylim([0, 1.1 * ymax])
    if scaleFactor == 1000:
        plt.ylabel('Photon count rate [kHz]')
    else:
        plt.ylabel('Photon count rate [Hz]')
    if len(list_of_channels) < 10:
        plt.legend(leg)
    plt.rcParams.update({'font.size': 20})
    plt.tight_layout()
    plt.yscale(yscale)

    return h


import numpy as np
from ..tools.find_nearest import find_nearest
from scipy.ndimage.filters import gaussian_filter

def sig(x, x0, a):
 return 1/(1 + np.exp(-a*(x-x0)))


def g2polar(g, Nr=180):
    """
    Function that converts x-correlations into a polar plot to indicate anisotropy
    
    Parameters
    ----------
    g : np.array()
        Array with 4 or 6 columns, each one the G values for right, up, left, down
        cross-correlations (only y values).
    Nr : int, optional
        Number of pixels in the polar plot. The default is 180.

    Returns
    -------
    z : np.array()
        2D array with the polar plot.

    """
    
    x = np.linspace(-1, 1, Nr)
    
    xx, yy = np.meshgrid(x, x)
    yy *= -1
    rr = np.sqrt(yy**2 + xx**2)
    theta = np.arctan(yy/xx)
    theta += 2*np.pi
    theta = theta % (np.pi)
    theta[yy < 0] += np.pi
    
    z = rr*0
    
    g_shape = np.shape(g)
    num_curves = g_shape[1]
    cmax = g_shape[0] - 1
    
    if num_curves == 4:
        angles = np.asarray([0, np.pi/2, np.pi, 3/2*np.pi, 2*np.pi])
    elif num_curves == 6:
        ['Right', 'UpRight', 'UpLeft', 'Left', 'DownLeft', 'DownRight']
        angles = np.asarray([0, np.pi/3, 2/3*np.pi, np.pi, 4/3*np.pi, 5/3*np.pi, 2*np.pi])
    else:
        return z
    
    for x in range(Nr):
        for y in range(Nr):
            r = np.clip(rr[y, x], 0, 1)
            th = theta[y, x]
            _, idx = find_nearest(angles, th)
            idx = np.mod(idx, len(angles)-1)
            color = g[np.clip(int(r*len(g[:,1])), 0, cmax), idx]
            z[y, x] = color
    
    z[rr > 1] = None
    return z

def g2flow(g, Nr=180, detector='square'):
    """
    Function that converts x-correlations into a polar plot to indicate flow
    
    Parameters
    ----------
    g : np.array()
        Array with 4 rows, each one the G values for right, up, left, down
        cross-correlations (only y values).
    Nr : int, optional
        Number of pixels in the polar plot. The default is 180.
    detector : str, optional
        Detector type used, either 'square' or 'airy' for the Zeiss airyscan

    Returns
    -------
    z : np.array()
        2D array with the flow plot.
    flow : list
        List with 2 values indicating the flow in tge [up, right] direction.

    """
    
    x = np.linspace(-1, 1, Nr)
    
    xx, yy = np.meshgrid(x, x)
    yy *= -1
    rr = np.sqrt(yy**2 + xx**2)
    theta = np.arctan(yy/xx)
    if detector == 'airy':
        # all angles shift with pi/4
        theta -= np.pi / 4
    theta += 2*np.pi
    theta = theta % (np.pi)
    theta[yy < 0] += np.pi
    
    z = rr*0
    
    G_diff = np.zeros((len(g[:, 0]), 2))
    G_diff[:, 0] = (g[:,1] - g[:,3]) / np.mean(g)
    G_diff[:, 1] = (g[:,0] - g[:,2]) / np.mean(g)
    cmax = len(G_diff[:,0]) - 1
    
    for x in range(Nr):
        for y in range(Nr):
            r = np.clip(rr[y, x], 0, 1)
            th = theta[y, x]
            color = G_diff[np.clip(int(r*len(G_diff[:,1])), 0, cmax), 1]*np.cos(th)
            color += G_diff[np.clip(int(r*len(G_diff[:,0])), 0, cmax), 0]*np.sin(th)
            z[y, x] = color
    
    #z = gaussian_filter(z, sigma=sigma)
    z[rr > 1] = None
    flow = np.sum(G_diff, 0)
    return z, flow


def g2polar_old(g, smoothing=3, norm=None):
    """
    Convert fcs correlation curves to polar flow heatmap
    
    Parameters
    ----------
    g : 1D np.array()
        Array with G values.
    smoothing : int
        Number indicating moving average window for smoothing

    Returns
    -------
    z : 1D np.array()
        Array with z values.
    
    """
    # smooth function
    Gsmooth = np.convolve(g, np.ones(smoothing)/smoothing, mode='valid')
    
    # calculate difference
    Gsmooth /= Gsmooth[0]
    Gdiff = Gsmooth - Gsmooth[0]
    Gout = Gdiff[1:] - Gdiff[0:-1]
    
    # smooth
    Gout = np.convolve(Gout, np.ones(10)/10, mode='valid')
    
    Gout -= np.min(Gout)
    Gout2 = []
    Gout2.append(Gout[0])
    offset = 0
    for i in range(len(Gout)-1):
        
        if Gout[i+1] > Gout[i]:
            offset += Gout[i+1] - Gout[i]
            Gout2.append(Gout[i+1]-offset)
        else:
            Gout2.append(Gout[i+1]-offset)
        
    return Gout2
    
    
    # set
    mask = np.clip(Gout[1:] - Gout[0:-1], None, 0)
    mask[mask < 0] = 1
    
    diff = np.diff(Gout)  # Calculate differences between consecutive elements
    mask2 = diff >= 0  
    Gout[1:][mask2] = Gout[:-1][mask2]
    
    
    
    
    Gout[1:] = Gout[1:]*mask
    
    Gout = np.convolve(Gout, np.ones(smoothing)/smoothing, mode='valid')
    
    return Gout
    
    # now we have the radius as a function of the color --> invert this
    rad = np.linspace(0, 1, 255)
    z = np.zeros(len(rad))
    
    for ri, r in enumerate(rad):
        z[ri] = Gdiff[np.clip(int(r*len(Gdiff)), 0, len(Gdiff)-1)]
    
    z += 1
    #z *= 1 - sig(rad,0.5,1)
    return np.cumsum(z - (1 - sig(rad,0.5,10))) * (1-rad)
        
    
    # calculate derivative
    Gdiff = np.clip(Gsmooth[1:] - Gsmooth[0:-1], a_min=None, a_max=0)
    Gdiff *= -1
    
    # normalized cumulative sum
    #Gdcum = np.cumsum(Gdiff)
    #Gdcum /= np.max(Gdcum)
    Gdcum = np.cumsum(Gsmooth)
    
    if norm is None:
        Gdcum /= np.max(Gdcum)
    else:
        Gdcum /= norm
    
    # now we have the radius as a function of the color --> invert this
    rad = np.linspace(0, 1, 255)
    z = np.zeros(len(rad))
    
    for ri, r in enumerate(rad):
        [dummy, idx] = find_nearest(Gdcum, r)
        z[ri] = 1 - idx / len(Gdcum)
    
    # apply sigmoid function to be more sensitive for changes near the center
    #z = sig(z, len(Gdcum)/1.5, 25/len(Gdcum))
    
    z = np.convolve(z, np.ones(7)/7, mode='valid')
    
    return z


from ..fcs.get_fcs_info import get_file_info, get_metafile_from_file
from .analysis_settings import FFSmetadata

def load_ffs_metadata(fname):
    """
    Load FFS metadata. Function assumes that metadata is stored in 
    'ffs_info.txt' with ffs the raw file name  (i.e. "fname" without .bin), e.g.
    'myFCSmeasurement.bin' is the raw data file, and
    'myFCSmeasurement_info.txt' is the metadata file
    OR
    metadata is stored in same .h5 file as the actual data
    ===========================================================================
    Input       Meaning
    ---------------------------------------------------------------------------
    ffs         File name of raw data file (not metadata file).
    ===========================================================================
    Output      Meaning
    ---------------------------------------------------------------------------
    metadata    analysis_settings metadata object with all fields filled in
                returns None if the file is not found
    ===========================================================================
    """
    
    # metadata
    fname = get_metafile_from_file(fname)
    md = get_file_info(fname)
    if md is not None:
        metadata = FFSmetadata()
        metadata_keys = [
            'num_pixels',
            'num_lines',
            'num_frames',
            'range_x',
            'range_y',
            'range_z',
            'num_datapoints',
            'hold_off_x5',
            'hold_off',
            'time_resolution',
            'dwelltime',
            'duration',
            'pxsize',]
        md_keys = [
            'numberOfPixels',
            'numberOfLines',
            'numberOfFrames',
            'rangeX',
            'rangeY',
            'rangeZ',
            'numberOfDataPoints',
            'holdOffx5',
            'holdOff',
            'timeResolution',
            'dwellTime',
            'duration',
            'pxsize',]
        for i, metadata_key in enumerate(metadata_keys):
            if md_keys[i] in list(md.__dict__.keys()) and getattr(md, md_keys[i]) is not None:
                setattr(metadata, metadata_key, getattr(md, md_keys[i]))
            else:
                setattr(metadata, metadata_key, 0)
        # get coordinates from file name
        # Cells_DEKegfp_4_LP70_75x75um_1500x1500px_y_52_x_1480.bin
        # returns [52, 1480]
        idx = fname.rfind("y_") + 2
        idx2 = fname.rfind("_x_")
        if fname[idx2+3:].rfind("_"):
            idx3 = fname.rfind("_")
        else:
            idx3 = fname.rfind(".")
        try:
            y = int(fname[idx:idx2])
        except:
            y = 0
        try:
            x = int(fname[idx2+3:idx3])
        except:
            x = 0
        metadata.coords = [y, x]
    else:
        metadata = None
    return metadata



import numpy as np


def spad_shifts():
    """

    Returns
    -------
    shift : 1D numpy array
        list of distances between SPAD elemenents (in a.u.).

    """
    
    shifts = []
    for xx in range(9):
        for yy in range(9):
            shifts.append(np.sqrt((xx-4)**2 + (yy-4)**2))
            
    shift = np.asarray(shifts)
    
    return shift


import numpy as np

def filter_g_obj(G, filt='sum5', f_acc=0.66):
    [indsorted, _, _] = sort_g_obj(G, filt)
    if indsorted is None:
        return None
    n_chunks = len(indsorted)
    chunks_on = np.zeros((n_chunks))
    indsorted = indsorted[0:int(np.round(f_acc*n_chunks))]
    chunks_on[indsorted] = 1
    return chunks_on
    

def sort_g_obj(G, filt):
    keylist = list(G.__dict__.keys())
    good_keys = []
    for ind, key in enumerate(keylist):
        if key.startswith(filt + '_chunk'):
            good_keys.append(key)
    
    if len(good_keys) == 0:
        return None, None, None
    
    g_shape = np.shape(getattr(G, good_keys[0]))
    Garray = np.zeros((g_shape[0],len(good_keys)))
    for i in range(len(good_keys)):
        Garray[:,i] = getattr(G, filt + '_chunk' + str(i))[:,1]
    
    [indsorted, Gsorted, dGmArray] = automatic_suppression_g(Garray)
    
    return indsorted, Gsorted, dGmArray
    

def automatic_suppression_g(G):
    """
    Automated suppression of artifacts in FCS data
    based on the article by Ries et al., Opt. Express, 2010.
    
    Parameters:
    G : np.ndarray
        2D array (m x n) with each column containing an autocorrelation curve
        (m data points, n curves).
        
    Returns:
    indsorted : np.ndarray
        List of the image indices from best to worst image.
    Gsorted : np.ndarray
        Sorted autocorrelation array from low to high deviation.
    dGmArray : np.ndarray
        Sorted autocorrelation deviations from low to high.
    
    """
    
    Gsize = G.shape
    n = Gsize[1]  # number of autocorrelation curves
    ind = np.arange(n)  # keeps track of the indices of the images
    indsorted = np.zeros(n, dtype=int)
    Gsorted = np.zeros(Gsize)
    dGmArray = np.zeros(n)
    
    if n == 2:
        Gsorted = G
        return indsorted, Gsorted, dGmArray
    
    for step in range(n-1, 0, -1):
        dG = np.zeros(step)
        
        # Calculate deviation for each curve
        for k in range(step):
            g_mean = np.mean(np.delete(G, k, axis=1), axis=1)
            dG[k] = np.mean((G[1:, k] - g_mean[1:]) ** 2)
        
        dGm = np.max(dG)
        m = np.argmax(dG)
        
        dGmArray[step] = dGm
        Gsorted[:, step] = G[:, m]
        indsorted[step] = ind[m]
        
        # Remove the worst autocorrelation function
        G = np.delete(G, m, axis=1)
        ind = np.delete(ind, m)
    
    Gsorted[:, 0] = G[:, 0]
    indsorted[0] = ind[0]

    return indsorted, Gsorted, dGmArray
